import turtle
from mpmath import mp
from collections import Counter
import matplotlib.pyplot as plt

# Configuration constants
GRID_SIZE = 50
TOTAL_CELLS = GRID_SIZE * GRID_SIZE
GRID_WIDTH = 600
LEGEND_WIDTH = 200
WINDOW_WIDTH = GRID_WIDTH + LEGEND_WIDTH + 50
WINDOW_HEIGHT = 700
CELL_SIZE = GRID_WIDTH / GRID_SIZE
SQUARE_SIZE = 30
SPACING = 32
LEGEND_HEIGHT = 380

# Generate π digits
mp.dps = TOTAL_CELLS + 10
pi_str = str(mp.pi)[2:2 + TOTAL_CELLS]
digits = [int(ch) for ch in pi_str]

# Define color palette
palette = {
    0: (0, 0, 0),        # Black
    1: (31, 119, 180),   # Blue
    2: (255, 127, 14),   # Orange
    3: (44, 160, 44),    # Green
    4: (214, 39, 40),    # Red
    5: (148, 103, 189),  # Purple
    6: (140, 86, 75),    # Brown
    7: (227, 119, 194),  # Pink
    8: (127, 127, 127),  # Gray
    9: (188, 189, 34)    # Yellow-green
}

# Convert RGB tuples to proportions (0-1) for turtle
def convert_rgb_to_proportions(palette):
    return {key: (r/255, g/255, b/255) for key, (r, g, b) in palette.items()}

palette_proportions = convert_rgb_to_proportions(palette)

def draw_grid(pen, screen, start_x, start_y, digits, palette, cell_size):
    for row in range(GRID_SIZE):
        for col in range(GRID_SIZE):
            idx = row * GRID_SIZE + col
            d = digits[idx]
            x = start_x + col * cell_size
            y = start_y - row * cell_size

            pen.penup()
            pen.goto(x, y)
            pen.pendown()
            pen.fillcolor(palette[d])
            pen.begin_fill()

            for _ in range(4):
                pen.forward(cell_size)
                pen.right(90)

            pen.end_fill()

        # Update display more frequently to show progress
        if row % 3 == 0:
            screen.update()

def draw_legend(pen, legend_x, legend_y, palette):
    # Create a white background with black border for the legend
    pen.penup()
    pen.color(0, 0, 0)  # Black border
    pen.fillcolor(1, 1, 1)  # White background

    # Draw legend background
    pen.goto(legend_x, legend_y + 50)  # Moved up to cover title
    pen.pendown()
    pen.width(2)  # Thicker border
    pen.begin_fill()
    for side in range(4):
        pen.forward(LEGEND_WIDTH if side % 2 == 0 else LEGEND_HEIGHT)
        pen.right(90)
    pen.end_fill()
    pen.width(1)  # Reset to normal width
    pen.penup()

    # Draw legend title centered
    title = "DIGIT COLOR LEGEND"
    title_x = legend_x + (LEGEND_WIDTH - len(title) * 8) // 2  # Center text, adjusted multiplier
    pen.goto(title_x, legend_y + 20)  # Moved up to fit in box
    pen.color(0, 0, 0)  # Black text
    pen.write(title, font=("Arial", 12, "bold"))

    # Draw color squares with proper alignment
    for d in range(10):
        y_off = legend_y - 20 - d * SPACING  # Shifted up by 20px

        # Draw color square
        pen.goto(legend_x + 20, y_off)
        pen.fillcolor(palette[d])  # The color for digit d
        pen.pendown()
        pen.begin_fill()
        for _ in range(4):
            pen.forward(SQUARE_SIZE)
            pen.right(90)
        pen.end_fill()
        pen.penup()

        # Write digit with larger font and better alignment
        pen.goto(legend_x + SQUARE_SIZE + 30, y_off + 5)  # Better text alignment
        pen.write(f"Digit {d}", font=("Arial", 12, "normal"))  # Correctly match color with digit

def plot_histogram(digits, palette):
    freq = Counter(digits)
    x_vals = sorted(freq.keys())
    y_vals = [freq[x] for x in x_vals]

    # Find digit with highest and lowest frequencies
    max_digit = max(freq.items(), key=lambda x: x[1])[0]
    min_digit = min(freq.items(), key=lambda x: x[1])[0]

    # Convert RGB tuples back to hex for matplotlib
    hex_palette = {key: f"#{format(int(r * 255), '02x')}{format(int(g * 255), '02x')}{format(int(b * 255), '02x')}"
                   for key, (r, g, b) in palette.items()}

    plt.figure(figsize=(8, 6))

    # Plot bars
    bars = plt.bar(x_vals, y_vals, color=[hex_palette[x] for x in x_vals])

    # Highlight max and min
    bars[max_digit].set_edgecolor('gold')
    bars[max_digit].set_linewidth(2)

    bars[min_digit].set_edgecolor('red')
    bars[min_digit].set_linewidth(2)

    # Add annotations for max and min
    max_count = freq[max_digit]
    min_count = freq[min_digit]

    plt.annotate(f'Highest: Digit {max_digit} ({max_count})',
                 xy=(max_digit, max_count),
                 xytext=(max_digit, max_count + 10),
                 fontsize=12,
                 ha='center',
                 va='bottom',
                 bbox=dict(boxstyle='round,pad=0.5', facecolor='gold', alpha=0.5))

    plt.annotate(f'Lowest: Digit {min_digit} ({min_count})',
                 xy=(min_digit, min_count),
                 xytext=(min_digit, min_count + 10),
                 fontsize=12,
                 ha='center',
                 va='bottom',
                 bbox=dict(boxstyle='round,pad=0.5', facecolor='red', alpha=0.2))

    plt.title("Digit Frequency in 50×50 Pi Grid", fontsize=14)
    plt.xlabel("Digit", fontsize=12)
    plt.ylabel("Count", fontsize=12)
    plt.xticks(x_vals)
    plt.tight_layout()
    plt.grid(axis='y', alpha=0.3)
    plt.show()

def main():
    try:
        # Set up Turtle with optimized window size
        screen = turtle.Screen()
        screen.title("Annotated Pi-Pixel Mosaic")
        screen.setup(width=WINDOW_WIDTH, height=WINDOW_HEIGHT)
        screen.bgcolor("white")  # White background for better contrast
        screen.tracer(0)  # Turn off animation for speed
        screen.colormode(1.0)  # Use 0-1 RGB values

        pen = turtle.Turtle()
        pen.hideturtle()
        pen.penup()
        pen.speed(0)  # Fastest speed

        # Create a background for the grid area
        grid_width = CELL_SIZE * GRID_SIZE
        grid_height = CELL_SIZE * GRID_SIZE
        start_x = -grid_width/2 + 20  # Center grid with a small offset
        start_y = grid_height/2 - 20  # Center grid with a small offset

        # Draw grid container with border
        pen.penup()
        pen.goto(start_x - 5, start_y + 5)
        pen.pendown()
        pen.color(0, 0, 0)  # Black border
        for _ in range(4):
            pen.forward(grid_width + 10 if _ % 2 == 0 else grid_height + 10)
            pen.right(90)
        pen.penup()

        # Add a small title for the grid
        pen.goto(start_x, start_y + 15)
        pen.write("Pi Digits Visualization (first 2500 digits)", font=("Arial", 10, "normal"))

        # Draw grid
        draw_grid(pen, screen, start_x, start_y, digits, palette_proportions, CELL_SIZE)

        # Draw legend
        grid_end_x = start_x + GRID_SIZE * CELL_SIZE + 20  # 20px margin after grid
        legend_x = grid_end_x
        legend_y = 250  # Start from upper part of the window
        draw_legend(pen, legend_x, legend_y, palette_proportions)

        screen.update()  # Final update to ensure everything is displayed

        print("Pi visualization complete! Close the window to continue.")
        turtle.mainloop()  # This is a better way to keep the window open

    except Exception as e:
        print(f"Error in turtle graphics: {e}")
    finally:
        # Even if turtle has issues, we can still show the histogram
        try:
            plot_histogram(digits, palette_proportions)
        except Exception as e:
            print(f"Error in plotting histogram: {e}")

if __name__ == "__main__":
    main()
